from typing import Any
from .mongobd import MongoDB
from bson.objectid import ObjectId
from pymongo.results import UpdateResult, DeleteResult

class SystemDB(MongoDB):

    def __init__(self, properties: dict, db: str, collection: str = "systems") -> None:
        super().__init__(properties, db)
        self.collection = collection

    def find_by_id(self, id: str) -> 'dict':

        _id = ObjectId(id)
        found = self.find_one({
            "_id": _id
        })

        if found is not None:
            found["id"] = str(found.pop("_id"))

        return found 

    def find_by_hash(self, hash: 'str', **kargs) -> 'None|dict':

        return self.find_one({"hash": hash}, **kargs)

    def insert_one(self, value: 'dict', **kargs) -> 'dict':

        if "id" in value.keys():
            value["_id"] = ObjectId(value.pop("id"))

        inserted_id = super().insert_one(self.collection, value, **kargs)

        return self.find_one({"_id": inserted_id})

    def update_one(self, id: 'str', update: 'dict', **kargs) -> 'dict':

        _id = ObjectId(id)
        updated = super().update_one(self.collection, {"_id": _id}, update, **kargs)

        updated["id"] = str(updated.pop("_id"))

        return updated

    def find_one(self, filter: 'Any', **kargs) -> 'None|dict':

        if "id" in filter.values():
            filter["_id"] = ObjectId(filter.pop("id"))

        found = super().find_one(self.collection, filter, **kargs)

        if found is None:
            return None

        found["id"] = str(found.pop("_id"))

        return found
    
    def delete_one(self, filter: 'Any', **kargs) -> 'None':
        
        if "id" in filter.keys():
            filter["_id"] = ObjectId(filter.pop("id"))

        super().delete_one(self.collection, filter, **kargs)
        

class UserDB(MongoDB):

    def __init__(self, properties: dict, db: str, collection: str = "users") -> None:
        super().__init__(properties, db)
        self.collection = collection
         
    def find_by_id(self, id: str) -> 'dict':
        
        _id = ObjectId(id)
        found = self.find_one({
            "_id": _id
        })

        if found is not None:
            found["id"] = str(found.pop("_id"))

        return found

    def find_many(self, filter: 'Any', **kargs) -> 'list[dict]|None':

        if "id" in filter.keys():
            filter["_id"] = ObjectId(filter.pop("id"))

        founds = super().find(self.collection, filter, **kargs)

        return founds

    def insert_one(self, value: 'dict', **kargs) -> 'dict|None':

        if "id" in value.keys():
            value["_id"] = ObjectId(value.pop("id"))

        inserted_id = super().insert_one(self.collection, value, **kargs)

        return  self.find_one({"_id": inserted_id})

    def update_one(self, id: 'str', update: 'dict', **kargs) -> 'dict|None':

        _id = ObjectId(id)

        updated = super().update_one(self.collection, {"_id": _id}, update, **kargs)
        updated["id"] = str(updated.pop("_id"))

        return updated

    def find_one(self, filter: 'Any', **kargs) -> 'UpdateResult':
        if "id" in filter.keys():
            filter["_id"] = ObjectId(filter.pop("id"))

        found = super().find_one(self.collection, filter, **kargs)
        found["id"] = str(found.pop("_id"))

        return found

    def delete_one(self, filter: 'Any', **kargs) -> 'DeleteResult':
        
        if "id" in filter.keys():
            filter["_id"] = ObjectId(filter.pop("id"))

        super().delete_one(self.collection, filter, **kargs)



class StreamTemplateNotFound(Exception):
    def __init__(self, 
    path_file: str,
    message: str = "Html Stream template not found" 
    ) -> None:
        self.path_file = path_file
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return f"{self.message} -> {self.path_file}"


import asyncio
import logging
import threading
import time
import uuid
from asyncio import Queue
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Callable, Iterable, List, Optional, Tuple
from bson import ObjectId

import cv2
import numpy as np
from apscheduler.triggers.cron import CronTrigger
from data_access.visual_knowledge_db import SystemDB, UserDB
from pydantic import Field
from tzlocal import get_localzone
from mvfy.entities.visual_knowledge_entities import User
from utils import constants as const, index as utils

from mvfy.visual import func
from mvfy.visual.detector import Detector
from mvfy.visual.systems.image_generator import ImageGenerator
from mvfy.visual.receiver.receivers import Receiver
from mvfy.visual.streamer import Streamer

from . import errors


@dataclass
class VisualKnowledge():

    detector_knows: Optional[Detector] = None
    detector_unknows: Optional[Detector] = None
    receiver: Optional[Receiver] = None
    streamer: Optional[Streamer] = None
    type_service: Optional[str] = None
    db_properties: Optional[dict] = None
    db_name: Optional[str] = None
    max_descriptor_distance: Optional[float] = None
    min_date_knowledge: Optional[float] = None
    min_frequency: Optional[float] = 0.7
    resize_factor: Optional[float] = 0.25
    features: Optional[list] = Field(default_factory=list)
    type_system: Optional[str] = const.TYPE_SYSTEM["OPTIMIZED"]
    title: Optional[str] = str(uuid.uuid4())
    delay: int = 30
    batch_images: int = 30
    draw_label: bool = False

    """
    Main model builder

    constructor
    :Parameters:
        {String} type_service - type of the listen server.
        {Dict} db - proprties of bd see https://pymongo.readthedocs.io/en/stable/api/pymongo/mongo_client.html#pymongo.mongo_client.MongoClient.
        {String} db_name - name of db to be used.
        {Array} min_date_knowledge [min_date_knowledge=null] - minimum interval to determine a known user.
        {Number} min_frequency [min_frequency=0.7] - minimum frequency between days detectioned.
        {list} features [features=null] - characteristics that will be saved in each detection.
        {String} max_descriptor_distance [max_descriptor_distance=null] - max distance of diference between detections.
        {String} type_system [type_system=null] - type of system.
        {String} title [title=null] - title of system.

    :Returns:
        None.

    """

    def __post_init__(self):

        # properties
        self.id = None

        # more info
        self.date_format = const.DATE_FORMAT
        self.cron_reload = self.__get_cron_trigger()

        # DB
        self.db_systems = SystemDB(
            properties=self.db_properties, db=self.db_name, collection=const.COLLECTIONS["SYSTEMS"]
        )
        self.db_users = UserDB(
            properties=self.db_properties, db=self.db_name, collection=const.COLLECTIONS["USERS"]
        )

        self.db_users.users.drop()

        self.new_users: Queue = Queue()
        self.evaluate_users: Queue = Queue()

        self._thread_receiver = utils.run_async_in_thread(self.receiver.start())
        self._thread_system = utils.run_async_in_thread(self.start())
        self._thread_insert_new_users = utils.run_async_in_thread(self.add_new_users())
        self._thread_evaluate_users = utils.run_async_in_thread(self.evaluate_detections())


    async def __preload(self) -> None:
        """
        Load system and users if exist

        Search the system or saved it
        Search the users of the system
        """
        print("reloading system...")
        # found or add system
        system = await func.get_system(self.get_obj(), self.db_systems)
        if system is None:
            system = await func.insert_system(self.get_obj(), self.db_systems)
            if system is None:
                raise errors.SystemNotFoundError(str(self.db_systems))

        # insert system found in instance
        self.__insert_system(system)
        logging.info("system created")

        # get descriptors
        await self.insert_known_users()
        await self.insert_unknown_users()
        logging.info("face of users inserted")

    def __get_cron_trigger(self) -> CronTrigger:
        """get crontrigger of now every day

        Returns:
            CronTrigger: trigger of action
        """
        _date = datetime.now(tz=get_localzone())
        return CronTrigger(
            hour=_date.hour, minute=_date.minute, second=_date.second, timezone=get_localzone()
        )

    def __insert_system(self, system: dict) -> None:
        """Insert system inside actual instance

        Args:
            system (dict): values to be replaced in instance
        """
        self.id = system["id"]
        self.type_service = (
            system["type_service"] if system["type_service"] is not None else self.type_service
        )
        self.title = system["title"] if system["title"] is not None else self.title
        self.features = system["features"] if system["features"] is not None else self.features
        self.min_date_knowledge = (
            system["min_date_knowledge"]
            if system["min_date_knowledge"] is not None
            else self.min_date_knowledge
        )
        self.min_frequency = (
            system["min_frequency"] if system["min_frequency"] is not None else self.min_frequency
        )
        self.max_descriptor_distance = (
            system["max_descriptor_distance"]
            if system["max_descriptor_distance"] is not None
            else self.max_descriptor_distance
        )
        self.type_system = (
            system["type_system"] if system["type_system"] is not None else self.type_system
        )
        self.resize_factor = (
            system["resize_factor"] if system["resize_factor"] is not None else self.resize_factor
        )

    async def insert_known_users(self):

        users = await func.get_users(
            {"system_id": self.id, "knowledge": True}, db=self.db_users
        )

        if users is not None:
            authors, encodings = zip(*[[user.author, user.detection] for user in users])
            self.detector_knows.authors = np.array(authors)
            self.detector_knows.encodings = np.array(encodings)

    async def insert_unknown_users(self):

        users = await func.get_users(
            {"system_id": self.id, "knowledge": False}, db=self.db_users
        )

        if users is not None:
            authors, encodings = zip(*[[user.author, user.detection] for user in users])
            self.detector_unknows.authors = np.array(authors)
            self.detector_unknows.encodings = np.array(encodings)

    def set_conf(
        self,
        date_format: Optional[str] = None,
        draw_label: Optional[bool] = None,
        cron_reload: Optional[CronTrigger] = None,
    ) -> None:
        """_summary_

        :param date_format: _description_, defaults to None
        :type date_format: Optional[str], optional
        :param draw_label: _description_, defaults to None
        :type draw_label: Optional[bool], optional
        :param cron_reload: _description_, defaults to None
        :type cron_reload: Optional[CronTrigger], optional
        """

        # more info
        self.date_format = date_format if date_format is not None else self.date_format
        self.draw_label = draw_label if draw_label is not None else self.draw_label
        self.cron_reload = cron_reload if cron_reload is not None else self.cron_reload

    def get_obj(self) -> dict:
        """Get a dict of the attributes for this instance.

        Returns:
            dict: instance relevant features
        """
        return {
            "title": self.title,
            "type_service": self.type_service,
            "max_descriptor_distance": self.max_descriptor_distance,
            "min_date_knowledge": self.min_date_knowledge,
            "min_frequency": self.min_frequency,
            "features": self.features,
            "type_system": self.type_system,
            "resize_factor": self.resize_factor,
            "id": self.id,
        }

    async def start(self) -> None:

        await self.__preload()

        func.async_scheduler(job=self.__preload, trigger=self.cron_reload)
        
        print("Start - detection of users")

        if self.batch_images % 2 != 0:
                raise ValueError(f"invalid batch image size, {self.batch_images}")
        
        while True:
            
            images_batch = [self.receiver.get() for _ in range(self.batch_images)]

            try:
                images_to_process = images_batch[1::2]
                images_raw = images_batch[0::2]

                faces = await asyncio.gather(*[self.detector_unknows.get_encodings(img) for img in images_to_process], return_exceptions=True)
                batch_processed = await asyncio.gather(*[self.detect_type_user(img, face_prop) for img, face_prop in zip(images_to_process, faces)], return_exceptions=True)

                all_images = []

                for pair_images in zip(images_raw, batch_processed):
                    all_images.extend(pair_images)

                await self.streamer.save(all_images)

            except Exception as error:
                logging.error(f"Error processing image {error}")
                await self.streamer.save(images_batch)

    async def evaluate_detections(self) -> None:
        while True:
            try:
                if not self.evaluate_users.empty():
                    author = await self.evaluate_users.get()
                    user_evaludated = await self.evaluate_detection(author)
                    
                    if user_evaludated.knowledge is True:
                        self.detector_knows.authors = np.append(self.detector_knows.authors, [user_evaludated.author], axis=0)
                        self.detector_knows.encodings = np.append(self.detector_knows.encodings, np.array(user_evaludated.detection).reshape(1, -1), axis=0)

            except Exception as error:
                logging.error(f"error evaluating user, {error}")

    async def add_new_users(self) -> None:

        while True:
            try:
                if not self.new_users.empty():
                    new_user = await self.new_users.get()
                    await func.insert_user(new_user, self.db_users)

                    self.detector_unknows.authors = np.append(self.detector_unknows.authors, [new_user['author']], axis=0)
                    self.detector_unknows.encodings = np.append(self.detector_unknows.encodings, np.array(new_user['detection']).reshape(1, -1), axis=0)

            except Exception as error:
                logging.error(f"error inserting new user, {error}")

    async def save_new_unknown(self, encoding: np.ndarray, features: list) -> None:

        new_author = str(uuid.uuid4())

        try:
            await self.new_users.put(
                {
                    "system_id": self.id,
                    "detection": encoding.tolist(),
                    "features": features,
                    "author": new_author,
                    "init_date": datetime.now(),
                    "last_date": datetime.now(),
                    "knowledge": False,
                    "frequency": 0,
                }
            )

        except Exception as error:
            logging.error(f"Error to insert a new user, {error}")

    async def save_evaluate_detection(self, author: str) -> None:

        try:
            await self.evaluate_users.put(author)
        except Exception as error:
            logging.error(f"Error to insert author for evaluate, {error}")

    @func.loop_manager
    async def draw_frame(
        self,
        image: np.ndarray,
        draw_label: bool,
        name: str,
        location: Any,
        color: Tuple[int, int, int],
        loop: 'asyncio.AbstractEventLoop',
    ) -> np.ndarray:
        # WARNING: programmer don't do this is a bad practice

        (top, right, bottom, left) = location

        await loop.run_in_executor(None, lambda: cv2.rectangle(image, (left, top), (right, bottom), color, 1))

        if draw_label is True:
            cv2.rectangle(image, (left, bottom), (right, bottom + 20), color, cv2.FILLED)
            font = cv2.FONT_HERSHEY_DUPLEX
            cv2.putText(image, name, (left + 10, bottom + 18), font, 0.8, (255, 255, 255), 1)

        return image

    async def detect_type_user(self, img: np.array, face_properties: Tuple[list, list]) -> "np.array":
        """Compare faces in image and detect knows and unkowns.

        Args:
            img (np.array): img with faces

        Returns:
            np.array: img
        """

        for face_location, face_encoding in zip(*face_properties):
            
            # location = self.detector_unknows.enlarge_dimensions(self.detector_unknows.face_locations[idx])
            known_comparations, unknown_comparations = await asyncio.gather(self.detector_knows.compare(face_encoding), self.detector_unknows.compare(face_encoding))

            if np.any(known_comparations):
                
                img = await self.draw_frame(
                    image = img,
                    draw_label = self.draw_label,
                    name = "conocido",
                    location = face_location,
                    color=(0, 128, 0),
                )

                continue

            if np.any(unknown_comparations):
                previous_authors = self.detector_unknows.authors[:len(unknown_comparations)] 
                await self.save_evaluate_detection(str(previous_authors[unknown_comparations][0]))
            else:
                await self.save_new_unknown(encoding=face_encoding, features=[])

            img = await self.draw_frame(
                    image = img,
                    draw_label = self.draw_label,
                    name = "desconocido",
                    location = face_location,
                    color=(0, 0, 255),
                )
        
        return img

    async def evaluate_detection(self, author: str) -> dict:
        """Evaluate the user s detection.

        Args:
            author (dict): author to be evaluated

        Returns:
            dict: user modified
        """
        user = await func.find_user({"author": author}, db=self.db_users)

        if user is None:
            logging.error(f"Error user not found in BD, author: {author}")
            return {}

        prev_user = user
        diff_date = utils.get_date_diff_so_far(user.init_date, self.min_date_knowledge[1])

        if diff_date > self.min_date_knowledge[0] and user.frequency >= self.min_frequency:
            
            user.knowledge = True

        elif utils.get_date_diff_so_far(user.last_date, self.min_date_knowledge[1]) > 0:

            prev_days = utils.frequency(
                total=self.min_date_knowledge[0], percentage=1, value=self.frequency, invert=True
            )
            user.last_date = datetime.now()
            user.frequency = utils.frequency(self.min_date_knowledge[0], 1, prev_days + 1)

        return (
            await func.update_user({**user, "modified_on": datetime.now()}, self.db_users)
            if prev_user != user
            else user
        )



import os
from typing import Dict, Tuple
from .feature_flags import ENVIROMENT

BASE_PROJECT = os.path.abspath(
    os.path.join(__file__, "../../..")
) if ENVIROMENT == "DEV" else os.getcwd()

UNKNOWS_URL = os.path.join(BASE_PROJECT, 'unknows_url')
ACQUAINTANCES_URL = os.path.join(BASE_PROJECT, 'acquaintances_url')
MODELS_URL = os.path.join(BASE_PROJECT, 'src/mvfy/models')
CONFIG_URL = os.path.join(BASE_PROJECT, 'config')
PORT = 3000 if (v:=os.getenv("PORT")) is None else v
IMAGE_DIMENSIONS = (720, 480)

#static
COLLECTIONS: Dict[str, str] = {
    "USERS": "users",
    "SYSTEMS": "systems"
}

#system
ALLOWED_FEATURES: Dict[str, str] = {
    "AGE": "age",
    "GENDER": "gender",
    "RACE": "race",
    "EMOTION": "emotion"
}
TYPE_SYSTEM: Dict[str, str] = {
    "OPTIMIZED": "optimized",
    "PRECISE": "precise"
}
TYPE_SERVICE: Dict[str, str] = {
    "REMOTE": "remote",
    "LOCAL": "local"
}
ACTION: Dict[str, str] = {
    "INIT_SYSTEM": "INIT_SYSTEM",
    "SET_DETECTION": "SET_DETECTION",
}
REQUEST: Dict[str, str] = {
    "ERROR": "ERROR",
    "GET_MODEL_FEATURES": "GET_MODEL_FEATURES",
    "GET_INITIALIZED_SYSTEM": "GET_INITIALIZED_SYSTEM",
    "SEND_DETECTION_VALIDATED": "SEND_DETECTION_VALIDATED",
    "LOCAL_IMAGE_SEND": "LOCAL_IMAGE_SEND"
}

HTML_STREAMER: Dict[str, str] = {
    "URL": os.path.join(BASE_PROJECT, '/public/streamer.html'),
    "URL_TEMP": os.path.join(BASE_PROJECT, '/public/temp/_streamer.html'),
    "PROTOCOL_REPLACE": '<<<PROTOCOL>>>',
    "HOST_REPLACE": '<<<HOST>>>',
    "EMIT_REPLACE": '<<<EMIT>>>',
}

# time
DATE_FORMAT: str = "%d/%m/%Y %H:%M:%S"
        
def DAYS (quantity: int) -> Tuple[int, str]:
    quantity = int(quantity)
    if isinstance(quantity, int):
        return (quantity, "days")
    else:
        raise ValueError("type of the quantity days is invalid")
    

def WEEKS (quantity: int) -> Tuple[int, str]:
    quantity = int(quantity)
    if isinstance(quantity, int):
        return (quantity, "weeks")
    else:
        raise ValueError("type of the quantity weeks is invalid")
    
def MONTHS  (quantity: int) -> Tuple[int, str]:
    quantity = int(quantity)
    if isinstance(quantity, int):
        return (quantity, "months")
    else: 
        raise ValueError("type of the quantity months is invalid")
    


from abc import ABC, abstractmethod
import asyncio
from dataclasses import dataclass, field
import time
from typing import Any, Optional, Tuple

import cv2
import face_recognition
import numpy as np
from cv2 import Mat
from mvfy.entities.visual_knowledge_entities import User
from mvfy.visual.func import loop_manager
from utils import index as utils


@dataclass
class Detector(ABC):

    authors: list = field(default=np.empty((0)))
    encodings: list = field(default=np.empty((0, 128)))
    resize_factor: Optional[float] = 0.25

    def reduce_dimensions_image(self, image: Any) -> Mat:
        """
        Resizes the image to less dimensions"""
        return cv2.resize(
            image, 
            dsize = (0, 0), 
            fx = self.resize_factor, 
            fy = self.resize_factor)

    def enlarge_dimensions(self, location: Tuple[int, ...]) -> Tuple[int, ...]:
        """
        Enlarges the location to the image size"""

        return_size = 1 / self.resize_factor
        (top, right, bottom, left) = location

        # Scale back up face locations since the frame we detected in was scaled to 1/4 size
        top *= return_size
        right *= return_size
        bottom *= return_size
        left *= return_size

        return tuple(map(int, (top, right, bottom, left))) 
    
@dataclass
class DetectorFaces(Detector):

    tolerance_comparation: float = 0.3

    @loop_manager
    async def get_encodings(self, image: Mat, loop: 'asyncio.AbstractEventLoop') -> Tuple[list, list]:
        """Detect encodings of faces in image

        Args:
            image (Mat): image with faces to compare

        Returns:
            List: List of 128-dimensional face encodings
        """
        face_encodings, face_locations = [], []
        image = await loop.run_in_executor(None, self.reduce_dimensions_image, image)

        face_locations = await loop.run_in_executor(None, face_recognition.face_locations, image) 

        if not (face_locations is None and face_locations != []):
            face_encodings = await loop.run_in_executor(None, face_recognition.face_encodings, image, face_locations)

        face_locations_resized = await loop.run_in_executor(None, lambda: list(map(self.enlarge_dimensions, face_locations)))

        return face_locations_resized, face_encodings

    @loop_manager
    async def compare(self, encoding: np.ndarray, loop: 'asyncio.AbstractEventLoop') -> list[bool]:
        res = await loop.run_in_executor(None, lambda: face_recognition.compare_faces(self.encodings, encoding, tolerance=self.tolerance_comparation))
        return res
        # return face_recognition.face_distance(self.encodings, encoding)
    
class DetectionFacesCPU(Detector):

    actual_img: np.ndarray = np.array([])
    face_locations: list = field(default_factory=list)
    
    
    async def get_encodings(self, image: Mat) -> list:
        """Detect encodings of faces in image

        Args:
            image (Mat): image with faces to compare

        Returns:
            List: List of 128-dimensional face encodings
        """

        self.actual_img = image

        self.reduce_dimensions_image()
        self.face_locations = face_recognition.face_locations(self.actual_img)
        face_encodings = []

        if not (self.face_locations is None and self.face_locations != []):
            face_encodings = face_recognition.face_encodings(self.actual_img, self.face_locations)

        return face_encodings

    async def compare(self, encoding: np.ndarray) -> np.ndarray:

        return face_recognition.face_distance(self.encodings, encoding)

    



from abc import ABC, abstractmethod
from asyncio import Queue
from typing import Any, Optional, Tuple

import cv2
import numpy as np
from pydantic.dataclasses import dataclass


@dataclass
class ImageGenerator(ABC):

    dimensions: Tuple[int, int] = (720, 480)
    wait_message: str = "wait...."
    wait_image: Any = None
    images_queue: Any = None
    images_queue_size: int = 0

    class Config:
        arbitrary_types_allowed = True

    def create_wait_image(self) -> None:
        """_summary_
        """
        self.wait_image = np.zeros([self.dimensions[1], self.dimensions[0], 1], dtype = np.uint8)
        center_image = (self.wait_image.shape[1] // 2, self.wait_image.shape[0] // 2)
        self.wait_image = cv2.putText(self.wait_image, self.wait_message, center_image, cv2.FONT_HERSHEY_SIMPLEX, 2, 255)

    def put_image(self, image: 'Any') -> None:
        """ put image inside queue """
        self.images_queue.put_nowait(image)
        self.images_queue_size += 1
    
    def get_image(self) -> Any:
        """get first image inside queue or wait image if empty"""
        
        if self.images_queue_size == 0:
            return self.wait_image

        _image = self.images_queue.get_nowait()
        self.images_queue_size -= 1

        return _image

    @abstractmethod
    def __iter__(self) -> None:
        pass



class FailedConnectionWithRSTP(Exception):
    """
    Exception raised when a connection to a remote device is failed.
    """
    pass
    def __init__(self, url_connection: str, *args: object) -> None:
        super().__init__(*args)
        self.url_connection = url_connection

    def __str__(self) -> str:
        return f"System can't connect with - {self.url_connection} -> {self.message}" 


from typing import Any
from pymongo import MongoClient
from pymongo.results import UpdateResult, BulkWriteResult, DeleteResult
from bson.objectid import ObjectId

class MongoDB:

    def __init__(self, properties: str, db: 'str') -> None:
        self.client = MongoClient(properties)
        self.url = self.client.address
        self.db = self.client[db]

    def __str__(self) -> str:
        return f"{self.url}"
        
    def insert_one(self, collection: str, value: 'Any', **kargs) -> 'ObjectId': 
        _collect = self.db[collection]

        _id = _collect.insert_one(value, **kargs).inserted_id
        return _id
    
    def insert_many(self, collection: str, value: 'Any', **kargs) -> 'list[ObjectId]': 
        _collect = self.db[collection]

        _ids = _collect.insert_many(value, **kargs).inserted_ids
        return _ids
    
    def replace_one(self, collection: str, registry_found: 'Any', registry_replace: 'Any', **kargs) -> 'UpdateResult': 
        _collect = self.db[collection]

        result = _collect.replace_one(registry_found, registry_replace, **kargs)
        return result
    
    def update_many(self, collection: str, filter: 'Any', update: 'Any', **kargs) -> 'list[UpdateResult]': 
        _collect = self.db[collection]

        result = _collect.update_many(filter, update, **kargs)
        return result
    
    def update_one(self, collection: str, filter: 'Any', update: 'Any', **kargs) -> 'UpdateResult': 
        _collect = self.db[collection]

        result = _collect.update_one(filter, update, **kargs)
        return result   
    
    def find(self, collection: str, filter: 'Any', **kargs) -> 'Any': 
        _collect = self.db[collection]

        result = _collect.find(filter, **kargs)
        return result

    def find_one(self, collection: str, filter: 'Any', **kargs) -> 'Any': 
        _collect = self.db[collection]

        result = _collect.find_one(filter, **kargs)
        return result
        
    def bulk_write(self, collection: str, requests: 'list[UpdateResult]', **kargs) -> 'BulkWriteResult':
        _collect = self.db[collection]

        result = _collect.bulk_write(requests, **kargs)
        return result

    def delete_one(self, collection: str, filter: 'Any', **kargs) -> 'DeleteResult': 
        _collect = self.db[collection]

        result = _collect.delete_one(filter, **kargs)
        return result
    
    def delete_many(self, collection: str, filter: 'Any', **kargs) -> 'list[DeleteResult]': 
        _collect = self.db[collection]

        result = _collect.delete_many(filter, **kargs)
        return result


from abc import ABC, abstractmethod
from asyncio import Queue
import asyncio
import time
from typing import Any, Iterable, Optional, Tuple, Union

import cv2
import numpy as np
from pydantic.dataclasses import dataclass
from . import errors
from mvfy.utils import constants


class Receiver(ABC):
    def __init__(self) -> None:
        pass
    
    @abstractmethod
    def start(self) -> None:
        pass
    
    @abstractmethod
    def get(self) -> bytes:
        pass

@dataclass
class ReceiverIpCam(Receiver):

    ip_cam: str
    dimensions: Tuple[int, int] = constants.IMAGE_DIMENSIONS
    stream: Any = None
    framerate: int = 30
    detections_failed: int = 0
    time_to_wait: int = 1
    images_queue: Optional[Any] = None
    images_queue_size: int = 0

    def __post_init__(self):
        
        self.images_queue = Queue()

    def __init_stream(self) -> None:

        print(f"conecting.... {self.ip_cam}")

        self.stream = cv2.VideoCapture(self.ip_cam)
        
        if self.stream is None:
            raise errors.FailedConnectionWithRSTP(self.ip_cam)
        
    async def start(self) -> None:

        try:
            self.__init_stream()

            while self.stream.isOpened():
                try:
                    ret, img = self.stream.read()
                    if ret:
                        await self.images_queue.put(img)
                        self.images_queue_size += 1
                        self.detections_failed = 0
                    else:
                        self.detections_failed += 1
                    if self.stream is None or self.detections_failed > self.framerate:
                        print(f"reconecting.... {self.ip_cam}")
                        self.__init_stream()

                except StopIteration:
                    print(f"reconecting.... {self.ip_cam}")
                    self.__init_stream()
                
                except Exception as error:
                    raise Exception(f"Error in stream connection {error}")

        except Exception as error:
            raise Exception(f"Error in connection to {self.ip_cam}, {error}")

    
    def get(self) -> np.array:

        # while self.images_queue.empty():
            # print('waiting ip cam ...')
        while self.images_queue.empty():
                time.sleep(time_to_wait)

        image = self.images_queue.get_nowait()
        self.images_queue_size -= 1

        return image
    
    def __iter__(self):

        return self

    def __next__(self):

        return self.get()
    
class ReceiverSocket(Receiver):
    """
    TODO: implement all properties to this class

    :param Receiver: _description_
    :type Receiver: _type_
    :raises Exception: _description_
    :raises Exception: _description_
    :return: _description_
    :rtype: _type_
    """
    server_socket: Union[tuple, str] = ""
    buffer_size: int = 1024

    def start(self):
        def inside_function():
            socketio = None
            sock = socketio.client()
            stream = None

            try:
                print(f"conecting.... {self.server_socket}")
                sock.connect(self.server_socket)
                print("init the capture of image")

                while True:
                    try:
                        if stream is None:
                            pass
                            # stream = socketio.on('connect')(lambda: yield stream.read())
                        
                        print(f"conecting....")

                    except Exception as err:
                        raise Exception(
                            f"Error in stream connection {err}")

            except Exception as err:
                raise Exception(
                    f"Error in conection to {err}")

        return inside_function


import asyncio
import logging
import uuid
from asyncio import AbstractEventLoop, Queue, Task
from datetime import datetime
from typing import Callable

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.cron import CronTrigger
import numpy as np
from data_access.visual_knowledge_db import SystemDB, UserDB
from entities.visual_knowledge_entities import System
from use_cases.visual_knowledge_cases import UserUseCases

from mvfy.use_cases.visual_knowledge_cases import SystemUseCases

from ..utils import index as utils


def async_scheduler(job: 'Callable', trigger: CronTrigger, *kargs) -> AsyncIOScheduler:
    """create a scheduler to be executed in future 

    :param job: _description_
    :type job: Callable
    :param trigger: _description_
    :type trigger: CronTrigger
    :param type: _description_, defaults to "cron"
    :type type: str, optional
    :return: _description_
    :rtype: _type_
    """
    _scheduler = AsyncIOScheduler()
    _scheduler.add_job(job, trigger = trigger, id = str(uuid.uuid4()), *kargs)
    _scheduler.start()

    return _scheduler
    
def loop_manager(func: 'Callable') -> 'Callable':
    """Decorator for Manage Event Loop.

    Args:
        func ([type]): callback async function
    
    Returns:
        (function): result of func
    """
    async def wrapper_function(*args, **kargs):
        loop = asyncio.get_event_loop()
        if loop is None:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        return await func(*args, **kargs, loop = loop)
        
    return wrapper_function

async def async_queue_object_put(list: list[dict], keys: list[str], queue: 'Queue' = Queue()) -> None:
    """Generate Queue with object properties extracted

    Args:
        list (list[dict]): list of objects 
        keys (list[str]): list of pairs key:value, required
        queue (Queue, optional): [description]. Defaults to Queue().
    
    Example
        >>> ... await async_queue_object([..., {..., "system": 1}], ["system"])

    """
    for obj in list:
        await queue.put({
            f"{k}":v for k, v in obj.items() if k in keys
        })
    
    queue.task_done()

async def async_queue_object_get(queue: 'Queue', callback: 'function', args: tuple = ()) -> None:

    while not queue.empty():
        res = await queue.get()
        await callback(*args, queue_result=res)

    queue.task_done()

@loop_manager
async def load_user_descriptors(system_id: str, db: UserDB, loop: 'asyncio.AbstractEventLoop') -> 'utils.ThreadedGenerator|None':
    """Load user descriptors from database.

    Args:
        system_id (str): id of parent system
        db (str): data-base of users
        loop (asyncio.AbstractEventLoop): actual event loop

    Returns:
        list[dict]: [description]
    """

    use_cases = UserUseCases(db)
    results = await loop.run_in_executor(None, lambda: use_cases.get_users({
        "system_id": system_id
    }))


    if results is None or results == []:
        results = []

    return utils.extract_objects(results, ["detection", "author"])

@loop_manager
async def get_system(system: 'dict', db: SystemDB, loop: 'asyncio.AbstractEventLoop') -> 'dict|None':
    """Get information about a system.

    Returns:
        dict: system found
    """
    use_cases = SystemUseCases(db)
    result = await loop.run_in_executor(None, lambda: use_cases.get_system(system))

    if result == [] or result is None:
        return None

    return result

@loop_manager
async def insert_system(system: 'dict', db: SystemDB, loop: 'asyncio.AbstractEventLoop') -> 'dict|None':
    """add information about a system.

    Returns:
        str: id of system insert
    """
    use_cases = SystemUseCases(db)
    result = await loop.run_in_executor(None, lambda: use_cases.add_system(system))

    if result == [] or result is None:
        return None

    return result

@loop_manager
async def insert_user(user: 'dict', db: UserDB, loop: 'asyncio.AbstractEventLoop') -> 'dict|None':
    """add information about a user.

    Returns:
        str: id of user insert
    """
    use_cases = UserUseCases(db)
    result = await loop.run_in_executor(None, lambda: use_cases.add_user(user))

    if result == [] or result is None:
        return None

    return result

@loop_manager
async def find_user(filter: 'dict', db: UserDB, loop: 'asyncio.AbstractEventLoop') -> 'dict|None':
    """search information about a user.

    Returns:
        str: id of user insert
    """
    use_cases = UserUseCases(db)
    result = await loop.run_in_executor(None, lambda: use_cases.get_user(filter))

    if result == [] or result is None:
        return None

    return result

@loop_manager
async def get_users(_filter: 'dict', db: UserDB, loop: 'asyncio.AbstractEventLoop') -> 'list':
    """search information about users.

    Returns:
        list: list of users
    """
    use_cases = UserUseCases(db)
    result = await loop.run_in_executor(None, lambda: use_cases.get_users(_filter))

    if result == [] or result is None:
        return None

    return result

@loop_manager
async def remove_users_duplicate_detection(_filter: 'dict', db: UserDB, loop: 'asyncio.AbstractEventLoop') -> None:
    """search information about users.

    Returns:
        list: list of users
    """
    use_cases = UserUseCases(db)
    users = await loop.run_in_executor(None, lambda: use_cases.get_users(_filter))

    if users != [] or not (users is None):
        detections = None
        for user in users:
            if detections is None:
                detections = np.array([user.detection])
            else:
                distance = np.apply_along_axis(lambda row: utils.euclidean_distance(row, np.array(user.detection)), 1, detections)


from typing import Any
from utils import index as utils
from entities.visual_knowledge_entities import System, User

class SystemUseCases:
    def __init__(self, db) -> None:
        self.db = db

    def add_system(self, system: 'System') -> 'None|dict':
        result = None
        if isinstance(system, (System)):
            result = self.db.insert_one(self.collection, system.get_obj())
        else:
            _obj = System(**system).get_obj()
            result = self.db.insert_one(_obj)
        
        return result
    
    def get_system(self, system: 'dict') -> 'None|dict':

        res_system = System(**system)

        if res_system.id is not None:
            result = self.db.find_by_id(res_system.id)
        else:
            result = self.db.find_by_hash(res_system.hash)
        return result
    
    def update_system(self, id: 'str', new_system: 'dict') -> 'None|dict':

        if id is None:
            raise ValueError("You must supply an id.")
        
        existing_system = self.db.find_by_id(id)
        if existing_system is None:
            return None
        
        updated_system = utils.distribute_object(existing_system, new_system)
        updated_system = System(**updated_system)

        return self.db.update_one(id, updated_system)
    
    def delete_system(self, filter: 'dict') -> 'None|dict':

        if filter["id"] is None:
            raise ValueError("You must supply an id.")
        
        return self.db.delete_one()

class UserUseCases:
    def __init__(self, db) -> None:
        self.db = db

    def add_user(self, user: 'User') -> 'None|dict':
        result = None
        if isinstance(user, (User)):
            result = self.db.insert_one(self.collection, user.get_obj())
        else:
            _obj = User(**user).get_obj()
            result = self.db.insert_one(_obj)
        
        return result
    
    def get_user(self, filter: 'dict') -> 'None|User':

        result = self.db.find_one(filter)
        res_user = User(**result)
        return res_user
    
    def get_users(self, filter: 'dict') -> 'None|list[User]':

        result = self.db.find_many(filter)

        if result is None:
            return []

        new_result = []
        for user in result:
            if '_id' in user.keys():
                user["id"] = str(user.pop("_id"))
                
            new_result.append(User(**user))

        return new_result

    def update_user(self, id: 'str', new_user: 'dict') -> 'None|Any':

        if id is None:
            raise ValueError("You must supply an id.")
        
        existing_user = self.db.find_by_id(id)
        if existing_user is None:
            return None
        
        updated_user = utils.distribute_object(existing_user, new_user)
        updated_user = User(**updated_user)

        return self.db.update_one(id, updated_user)
    
    def delete_user(self, filter: 'dict') -> 'None|Any':

        if filter["id"] is None:
            raise ValueError("You must supply an id.")
        
        return self.db.delete_one()


from datetime import datetime
from typing import Optional
from utils.constants import TYPE_SERVICE, TYPE_SYSTEM
import numpy as np
import hashlib

class MetaUser(type):

    def __call__(self, 
            system_id,
            author,
            detection,
            init_date,
            last_date,
            features,
            knowledge,
            frequency,
            *args,
            **kwds):

        if not isinstance(system_id, (str)):
            raise ValueError(
                f"Invalid system_id: {system_id}, must be a string")

        if not isinstance(author, (str)):
            raise ValueError(
                f"Invalid author: {author}, must be a string")

        if not isinstance(detection, list):
            raise ValueError(
                f"Invalid detection: {detection}, must be a np.float64")

        if not type(init_date) is datetime:
            raise ValueError(
                f"Invalid init_date: {init_date}, must be a datetime")

        if not type(last_date) is datetime:
            raise ValueError(
                f"Invalid last_date: {last_date}, must be a datetime")

        if not isinstance(features, (dict, object)) and features is not None:
            raise ValueError(
                f"Invalid features: {features}, must be a dictionary")

        if not isinstance(knowledge, (bool)) and knowledge is not None:
            raise ValueError(
                f"Invalid knowledge: {knowledge}, must be a dictionary")

        if not isinstance(frequency, (np.number, int, float)) or frequency < 0 or frequency > 1 or frequency is None:
            raise ValueError(
                f"Invalid frequency: {frequency}, must be a number, between 0 and 1")

        return super().__call__(
            system_id,
            author,
            detection,
            init_date,
            last_date,
            features,
            knowledge,
            frequency,
            *args,
            **kwds)


class MetaSystem(type):

    def __call__(self, 
            type_service,
            max_descriptor_distance,
            min_date_knowledge,
            min_frequency,
            resize_factor,
            features,
            type_system,
            title,
            id,
            *args,
            **kwds):

        if not isinstance(type_service, (str)) or type_service not in TYPE_SERVICE.values():
            raise ValueError(f"Invalid type_service: {type_service}")

        if not isinstance(max_descriptor_distance, (float)) or max_descriptor_distance <= 0 or max_descriptor_distance > 1:
            raise ValueError(
                f"Invalid max_descriptor_distance: {max_descriptor_distance}")

        if not isinstance(min_date_knowledge, (list, tuple)):
            raise ValueError(
                f"Invalid min_date_knowledge: {min_date_knowledge}")

        if not isinstance(min_frequency, (float)):
            raise ValueError(
                f"Invalid min_frequency: {min_frequency}, must be a float")

        if not isinstance(resize_factor, (float)):
            raise ValueError(
                f"Invalid resize_factor: {resize_factor}, must be a float")

        if not isinstance(features, (list)):
            raise ValueError(
                f"Invalid features: {features}, must be a list")

        if not isinstance(type_system, (str)) or type_system not in TYPE_SYSTEM.values():
            raise ValueError(
                f"Invalid type_system: {type_system}, must be a string or valid TYPE SYSTEM see insert link")

        if not isinstance(title, (str)):
            raise ValueError(
                f"Invalid title: {title}, must be a string")

        return super().__call__(type_service,
                 max_descriptor_distance,
                 min_date_knowledge,
                 min_frequency,
                 resize_factor,
                 features,
                 type_system,
                 title,
                 id)
        

class User(metaclass=MetaUser):
    """Entity User

    Args:
        metaclass (class, optional): Meta class of user. Defaults to MetaUser.
    """

    def __init__(self,
                 system_id: str,
                 author: str,
                 detection: list,
                 init_date: datetime,
                 last_date: datetime,
                 features: dict = {},
                 knowledge: bool = False,
                 frequency: int = 0,
                 created_on: datetime = datetime.now(),
                 modified_on: datetime = datetime.now(),
                 id: str = None,
                 ) -> None:

        self.detection = detection
        self.features = features
        self.init_date = init_date
        self.last_date = last_date
        self.knowledge = knowledge
        self.frequency = frequency
        self.author = author
        self.created_on = created_on
        self.modified_on = modified_on
        self.system_id = system_id
        self.id = id

    def get_obj(self) -> dict:
        return {
            "detection": self.detection,
            "features": self.features,
            "init_date": self.init_date,
            "last_date": self.last_date,
            "knowledge": self.knowledge,
            "frequency": self.frequency,
            "author": self.author,
            "created_on": self.created_on,
            "modified_on": self.modified_on,
            "system_id": self.system_id,
            "id": self.id
        }


class System(metaclass=MetaSystem):

    def __init__(self,
                 type_service: str,
                 max_descriptor_distance: float,
                 min_date_knowledge: list,
                 min_frequency: float,
                 resize_factor: float,
                 features: list,
                 type_system: str,
                 title: str,
                 created_on: datetime = datetime.now(),
                 modified_on: datetime = datetime.now(),
                 id: str = None,
                 ) -> None:

        self.type_service = type_service
        self.max_descriptor_distance = max_descriptor_distance
        self.min_date_knowledge = min_date_knowledge
        self.min_frequency = min_frequency
        self.resize_factor = resize_factor
        self.features = features
        self.type_system = type_system
        self.id = id
        self.title = title
        self.created_on = created_on
        self.modified_on = modified_on

    def get_obj(self) -> dict:
        return {
            "type_service": self. type_service,
            "max_descriptor_distance": self. max_descriptor_distance,
            "min_date_knowledge": self. min_date_knowledge,
            "min_frequency": self. min_frequency,
            "resize_factor": self.resize_factor,
            "features": self. features,
            "type_system": self. type_system,
            "id": self. id,
            "title": self. title,
            "hash": self.hash,
            "created_on": self.created_on,
            "modified_on": self.modified_on,
        }
    
    @property
    def hash(self) -> str:
        str2hash = f"{self.title}{self.type_system}"
        return hashlib.md5(str2hash.encode()).hexdigest()


import asyncio
import logging
import os
import time
from abc import ABC, abstractmethod
from asyncio import Queue
from typing import Any, Generator, Optional, Tuple

import cv2
from flask import render_template_string
import numpy as np
from mvfy.visual.func import loop_manager
from mvfy.visual.systems.image_generator import ImageGenerator
from pydantic.dataclasses import dataclass

from .errors import StreamTemplateNotFound


class Streamer(ABC):

    @abstractmethod
    def send(self, image: Any, execution_time: float)-> bytes:
        pass 

@dataclass
class FlaskStreamer(Streamer):

    dimensions: Tuple[int, int] = (720, 480)
    extension: Optional[str] = ".jpg"
    resize: Optional[Tuple[int, int]] = None
    images_queue: Optional[Any] = None
    images_queue_size: int = 0
    wait_message: str = "wait...."
    wait_image: Any = None
    framerate: int = 24
    end_time_return: float = time.time()

    def __post_init__(self):
        self.images_queue = Queue()
        self.create_wait_image()

    def get_template(self) -> str:
        """_summary_

        :raises StreamTemplateNotFound: _description_
        :return: _description_
        :rtype: str
        """        
        dir_name: str = os.path.dirname(os.path.abspath(__file__))
        template_path: str = os.path.join(dir_name, "stream_flask_template.html")
        
        if not os.path.exists(template_path):
            raise StreamTemplateNotFound(path_file = template_path)
        
        with open(template_path, "r", encoding = "utf-8") as f:
            template = f.read()

        return render_template_string(template, title = "mvfy_visual")

    def create_wait_image(self) -> None:
        """_summary_
        """
        self.wait_image = np.zeros([self.dimensions[1], self.dimensions[0], 1], dtype = np.uint8)
        center_image = (self.wait_image.shape[1] // 2, self.wait_image.shape[0] // 2)
        self.wait_image = cv2.putText(self.wait_image, self.wait_message, center_image, cv2.FONT_HERSHEY_SIMPLEX, 2, 255)

        flag, resize_image = cv2.imencode(self.extension, self.wait_image, [cv2.IMWRITE_JPEG_QUALITY, 80])
        if flag:
            self.wait_image: bytes = b'--frame\r\n' b'Content-Type: image/jpeg\r\n\r\n' + bytearray(resize_image) + b'\r\n'
    
    @loop_manager
    async def img2bytes(self, image, loop: 'asyncio.AbstractEventLoop') -> bytes:

        images_bytes = b'--frame\r\n' b'Content-Type: image/jpeg\r\n\r\n' + bytearray() + b'\r\n'
        flag, resize_image = await loop.run_in_executor(None, lambda: cv2.imencode(self.extension, image, [cv2.IMWRITE_JPEG_QUALITY, 80]))

        if flag:
            images_bytes: bytes = b'--frame\r\n' b'Content-Type: image/jpeg\r\n\r\n' + bytearray(resize_image) + b'\r\n'
        
        return images_bytes
    
    async def save(self, batch_images: list[Any]) -> Any:

            tasks = [self.img2bytes(img) for img in batch_images]
            results = await asyncio.gather(*tasks)

            for result in results:
                await self.images_queue.put(result)
                self.images_queue_size += 1

    def send(self)-> bytes:
        """_summary_

        :return: _description_
        :rtype: _type_
        """       
        #TODO: optimize the fluency of the video
        
        try:
            
            while self.images_queue.empty():
                time.sleep(0.1)

            image_to_send = self.images_queue.get_nowait()
            self.images_queue_size -= 1

            # wait = (1 / self.framerate) - (time.time() - self.end_time_return)
            # if wait < 0:
            #     print(f'delay:{wait} ')

            delay_time = max(0, (1 / self.framerate) - (time.time() - self.end_time_return))
            time.sleep(delay_time)

            self.end_time_return = time.time()

            return image_to_send
            
        except Exception as error:
            logging.error(f"Error sending the image, {error}")
            return self.wait_image
    
    def __iter__(self):

        while self.images_queue.empty():
            print("waiting streaming...")
            while self.images_queue.empty():
                time.sleep(1)
                
        return self

    def __next__(self):

        return self.send()
        
        




class SystemNotFoundError(Exception):
    def __init__(self, db: str, *args: object) -> None:
        super().__init__(*args)
        self.db = db

    def __str__(self) -> str:
        return f"System not found in DB or not it couldn't be saved - {self.db}"
    

 


from flask import Flask, Response
from flask_cors import CORS

from mvfy.utils import constants as const, index as utils
from mvfy.visual.detector.detectors import DetectorFaces
from mvfy.visual.systems import VisualKnowledge
from mvfy.visual.receiver import ReceiverIpCam
from mvfy.visual.streamer import FlaskStreamer

app = Flask(__name__)
CORS(app, resources={r"*": {"origins": "*"}})

receiver = ReceiverIpCam(ip_cam="rtsp://mvfy:mvfy@192.168.1.6:8080/h264_ulaw.sdp")
detector_knows = DetectorFaces(tolerance_comparation= 1 - 0.7)
detector_unknows = DetectorFaces(tolerance_comparation= 1 - 0.7)
streamer = FlaskStreamer()

visual = VisualKnowledge(
    detector_knows=detector_knows,
    detector_unknows=detector_unknows,
    receiver=receiver,
    streamer=streamer,
    type_service=const.TYPE_SERVICE["LOCAL"],
    db_properties="mongodb://localhost:27017/",  # type: ignore
    db_name="mvfy",
    max_descriptor_distance=0.6,
    min_date_knowledge=const.DAYS(7),
    type_system=const.TYPE_SYSTEM["OPTIMIZED"],
    features=[],
    title="mvfy_1",
    draw_label = False,
)


@app.route("/")
def index():
    return streamer.get_template()

@app.route("/stream_video")
def stream_video() -> Response:
    # image = streamer.send()
    return Response(streamer, mimetype="multipart/x-mixed-replace; boundary=frame")


app.run(host="0.0.0.0", port=8001, debug=False)




# -*- coding: utf-8 -*-

# Learn more: https://github.com/erwingforerocastro/mvfy_visual

from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()

with open('LICENSE.txt') as f:
    license = f.read()

classifiers = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Operating System :: Microsoft :: Windows',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.9'
]

setup(
    name='mvfy_visual',
    version='0.0.5',
    description='Package mvfy_visual',
    # long_description = readme,
    author='Erwing Forero',
    author_email='erwingforerocastro@gmail.com',
    url='https://github.com/erwingforerocastro/mvfy_visual',
    license=license,
    keywords = ['python', 'video', 'streaming', 'opencv', 'face recognition', 'mongodb'],
    classifiers=classifiers,
    py_modules=['mvfy'],
    package_dir={"": "mvfy"},
    packages=find_packages(where="mvfy", exclude=('test','docs'))
)



import asyncio
from datetime import datetime
import inspect
import logging
from queue import Queue
from threading import Thread
import threading
from typing import Any, Callable, Coroutine, Iterable

import numpy as np
from utils.constants import DATE_FORMAT

def euclidean_distance(x_1: np.array, x_2: np.array) -> np.float64:

    return np.sqrt(np.sum((x_1-x_2)**2))

def get_actual_date(format: str) -> str:
    """Get the actual date from a given format.

    Args:
        format (str): valid format datetime

    Returns:
        datetime: formatted datetime
    """
    date = datetime.now()
    try:
        return date.strftime(format) if format is not None else str(date)
    except Exception as e:
        logging.error(f"get_actual_date - Error to format datetime {e}")
        return str(date)
        
def distribute_object(object_left: dict, object_right: dict) -> dict:
    """Insert object_right inside object_left
    
    Expected: same keys in both dicts
    
    Args:
        object_left (dict): object to be inserted
        object_right (dict): object to insert

    Raises:
        ValueError: [description]

    Returns:
        dict: objects merged
    """
    temp_object = {}

    if get_diff_list(object_left.keys(), object_right.keys()) is not None or get_diff_list(object_left.keys(), object_right.keys()) != []:
        raise ValueError("Invalid keys between objects")

    for key, item in object_right.items():
        temp_object[key] = object_left[key] if item is None else item

    return temp_object

def get_diff_list(lists: 'tuple(list)',   _type: 'str' = 'all') -> list:
    """Get difference between two list

    Args:
        lists (tuple): two list to be compared
        _type (str, optional): _type of get diff:

        all - get all list values different
        left - get only left different values
        right - get only right different values
        
        Defaults to 'all'.

    Raises:
        ValueError: Invalid size of lists, expected: __len__ 2
        ValueError: Invalid _type of lists

    Returns:
        list: difference
    """
    if len(lists) != 2:
        raise ValueError("Invalid size of lists, expected: __len__ 2")

    if not is_iterable(lists[0]) or not is_iterable(lists[1]):
        raise ValueError("Invalid _type of lists")

    diff = list(set(lists[0]) ^ set(lists[1]))

    if _type == "left":
        diff = [column for column in diff if column in lists[0]]
    
    elif _type == "right":
        diff = [column for column in diff if column in lists[1]]

    elif _type == "left":
        pass
    
    return diff

def is_iterable(posibleList):
    """Validate if element is iterable

    Args:
        posibleList (Any): posible iterable element

    Returns:
        bool: if element is iterable
    """
    try:
        if isinstance(posibleList, (tuple, list)) or hasattr(posibleList, "__iter__"):
            _ = posibleList[0]
            return True

        return False
    except Exception as e:
        return False

def extract_objects(object_list: list[dict], keys: list[str]) -> list[dict]:
    """Extracts keys from a list of objects.

    Args:
        object_list (list[dict]): list of objects
        keys (list[str]): keys that be extracted

    Returns:
        list[dict]: list of objects
    """
    res = []
    object_list = object_list if is_iterable(object_list) else [object_list]

    for obj in object_list:
        res.append({
            f"{k}":v for k, v in obj.items() if k in keys
        })

    return res

def get_date_diff_so_far(date: datetime, _type: str = "days") -> float:
    """Returns the difference between two date objects.

    Args:
        date (datetime): date to be compared with today
        _type (str, optional): type of difference <"days" or "weeks" or "months">. Defaults to "days".

    Returns:
        float: difference between "date" and now
    """
    date_now = datetime.now()
    if date is not None:
        res = None
        if _type == "days":
            res = abs((date_now - date).days) 
        elif _type == "weeks":
            res = abs((date_now - date).days) // 7
        elif _type == "months":
            res = abs((date_now.year - date.year) * 12 + date_now.month - date.month)

        return res
    
def frequency(total: float, percentage: float, value: float, invert: bool = False) -> float:
    """Return frequency of a given value .

    Freq = v*t/p

    v-> value
    t-> total
    p-> percentage

    Args:
        total (float): total value
        percentage (float): percentage value between 0 and 1
        value (float): value
        invert (bool, optional): change total by percentage in formula. Defaults to False.

    Returns:
        float: frequency result
    """
    return (value * total) / percentage if invert is True else (value * percentage) / total

def run_async_in_thread(coroutine: Coroutine) -> Thread:
    """Run a coroutine in a thread.
    Args:
        coroutine (Coroutine): coroutine to run"""
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    
    thread = threading.Thread(target = loop.run_until_complete, args = (coroutine, ))
    thread.daemon = True
    thread.start()

    return thread

#------------------------------------------------------------------------------- CLASES ----------------------------------------------------------------------------------------

class ThreadedGenerator():
    """
    Generator that runs on a separate thread, returning values to calling
    thread. Care must be taken that the iterator does not mutate any shared
    variables referenced in the calling thread.
    """

    def __init__(self, 
        iterator: Iterable,
        daemon = True,
        Thread = Thread,
        sentinel: 'Any' = None,
        queue=Queue):

        self.size = len(iterator)
        self._iterator = iterator
        self._sentinel = sentinel
        self._queue = queue()
        self._thread = Thread(
            target=self._run
        )
        self._thread.daemon = daemon
        self._cb = None

    def insert_action(self, cb: 'Callable', args=()):
        """Add a callback to the intance.

        Args:
            cb (function): callback function
            args (tuple, optional): args. Defaults to ().

        """
        self._cb = lambda value: cb(value, *args)

    def __repr__(self):
        return f"ThreadedGenerator({self._iterator})"

    def _run(self):
        """Execute queue put process
        """
        for value in self._iterator:
            if self._cb is not None:
                try:
                    value = self._cb(value)
                except Exception as err:
                    logging.warning(f"{inspect.currentframe().f_code.co_name} Error inside function: \n {err}")

            self._queue.put(value)
            
        self._queue.put(self._sentinel)
        self._queue.task_done()

    def __iter__(self):
        """Iterate over the values in the queue .

        Yields:
            Any: value in the queue
        """
        self._thread.start()
        end = -1

        for value in iter(self._queue.get, self._sentinel):
            if value != self._sentinel:
                end += 1

            if end >= self.size:
                break

            yield value

        
        self._thread.join()

